# -*- coding: utf-8 -*-
import inspect
import re
from datetime import datetime

from hmac import new as hnew
from hashlib import sha256
from .signature_v3 import validate_v3_signature

try:
    from urllib.parse import urlparse, urlunparse
except ImportError:
    from urlparse import urlparse, urlunparse

try:
    from base64 import encodebytes as base64_encode
except ImportError:
    from base64 import encodestring as base64_encode

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    from inspect import getargspec as getargspec

def validate_signature(uri, nonce, signature, auth_token=''):
    """
    Validates requests made by Plivo to your servers.

    :param uri: Your server URL
    :param nonce: X-Plivo-Signature-V2-Nonce
    :param signature: X-Plivo-Signature-V2 header
    :param auth_token: Plivo Auth token
    :return: True if the request matches signature, False otherwise
    """

    auth_token = bytes(auth_token.encode('utf-8'))
    nonce = bytes(nonce.encode('utf-8'))
    signature = bytes(signature.encode('utf-8'))

    parsed_uri = urlparse(uri.encode('utf-8'))
    base_url = urlunparse((parsed_uri.scheme.decode('utf-8'),
                           parsed_uri.netloc.decode('utf-8'),
                           parsed_uri.path.decode('utf-8'), '', '',
                           '')).encode('utf-8')

    return base64_encode(hnew(auth_token, base_url + nonce, sha256)
                         .digest()).strip() == signature


def is_valid_time_comparison(time):
    if isinstance(time, datetime):
        return True
    return False


def is_valid_subaccount(subaccount):
    subaccount_string = str(subaccount)
    if len(subaccount_string) == 20 and subaccount_string[:2] == 'SA':
        return True
    return False


def is_valid_mainaccount(mainaccount):
    mainaccount_string = str(mainaccount)
    if len(mainaccount_string) == 20 and mainaccount_string[:2] == 'MA':
        return True
    return False


def to_param_dict(func, vals, exclude_none=True, func_args_check=True):
    args = getargspec(func)[0]
    arg_names = list(args)
    # The bit of regex magic below is for arguments that are keywords in
    # Python, like from. These can't be used directly, so our convention is to
    # add "_" suffixes to them. This strips them out.
    pd = {
        re.sub(r'^(.*)_+$', r'\1', key): value
        for key, value in vals.items()
        if key != 'self' and (key in arg_names or func_args_check==False) and (
            value is not None or exclude_none is False)
    }
    return pd



import six

from plivo.xml import (
    PlivoXMLElement,
    map_type,
    BreakElement,
    EmphasisElement,
    LangElement,
    PhonemeElement,
    ProsodyElement,
)


class SElement(PlivoXMLElement):
    _name = 's'
    _nestable = [
        'break',
        'emphasis',
        'lang',
        'phoneme',
        'prosody',
        'say-as',
        'sub',
        'w'
    ]

    def __init__(
        self,
        content=None,
    ):

        super(SElement, self).__init__()
        self.content = content

    def to_dict(self):
        d = {}
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }

    def add_break(
        self,
        strength=None,
        time=None
    ):

        self.add(
            BreakElement(
                strength=strength,
                time=time,
            ))
        return self

    def add_emphasis(
        self,
        content,
        level=None,
    ):

        self.add(
            EmphasisElement(
                content=content,
                level=level,
            ))
        return self

    def add_lang(
        self,
        content,
        xmllang=None,
    ):

        self.add(
            LangElement(
                content=content,
                xmllang=xmllang,
            ))
        return self

    def add_phoneme(
        self,
        content,
        alphabet=None,
        ph=None,
    ):

        self.add(
            PhonemeElement(
                content=content,
                alphabet=alphabet,
                ph=ph,
            ))
        return self

    def add_prosody(
        self,
        content,
        volume=None,
        rate=None,
        pitch=None,
    ):

        self.add(
            ProsodyElement(
                content=content,
                volume=volume,
                rate=rate,
                pitch=pitch,
            ))
        return self

    def add_say_as(
        self,
        content,
        interpret_as=None,
        format=None,
    ):
        from .sayAsElement import SayAsElement

        self.add(
            SayAsElement(
                content=content,
                interpret_as=interpret_as,
                format=format,
            ))
        return self

    def add_sub(
        self,
        content,
        alias=None,
    ):
        from .subElement import SubElement

        self.add(
            SubElement(
                content=content,
                alias=alias,
            ))
        return self

    def add_w(
        self,
        content,
        role=None,
    ):
        from .wElement import WElement

        self.add(
            WElement(
                content=content,
                role=role,
            ))
        return self



# -*- coding: utf-8 -*-
"""
Number & PhoneNumber classes - along with their list classes
"""

from plivo.base import (ListResponseObject, PlivoResource,
                        PlivoResourceInterface)
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Number(PlivoResource):
    _name = 'Number'
    _identifier_string = 'number'

    def delete(self):
        return self.client.numbers.delete(self.id)

    def update(self,
               app_id=None,
               subaccount=None,
               alias=None,
               verification_info=None):
        return self.client.numbers.update(self.id, app_id, subaccount, alias,
                                          verification_info)


class Numbers(PlivoResourceInterface):
    def __init__(self, client):
        self._resource_type = Number
        super(Numbers, self).__init__(client)

    @validate_args(
        number=[is_phonenumber()],
        app_id=[optional(of_type(six.text_type))],
        verification_info=[optional(of_type_exact(dict))])
    def buy(self, number, app_id=None, verification_info=None):
        return self.client.request('POST', ('PhoneNumber', number),
                                   to_param_dict(self.buy, locals()))

    def search(self,
               country_iso,
               type=None,
               pattern=None,
               region=None,
               services=None,
               lata=None,
               rate_center=None,
               limit=None,
               offset=None,
               eligible=None,
               city=None,
               npanxx=None,
               local_calling_area=None):
        return self.client.request('GET', ('PhoneNumber', ),
                                   to_param_dict(self.search, locals()))

    @validate_args(
        services=[
            optional(
                is_iterable(
                    all_of(of_type(six.text_type), is_in(('sms', 'voice', 'mms'))),
                    sep=','))
        ],
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self,
             type=None,
             number_startswith=None,
             subaccount=None,
             alias=None,
             services=None,
             tendlc_registration_status=None,
             tendlc_campaign_id=None,
             toll_free_sms_verification=None,
             limit=20,
             offset=0):
        return self.client.request(
            'GET',
            ('Number', ),
            to_param_dict(self.list, locals()),
            objects_type=Number,
            response_type=ListResponseObject, )

    @validate_args(number=[is_phonenumber()])
    def get(self, number):
        return self.client.request(
            'GET', ('Number', number), response_type=Number)

    @validate_args(
        numbers=[
            one_of(
                is_iterable(of_type(six.text_type), sep=','),
                of_type(six.text_type))
        ],
        carrier=[of_type(six.text_type)],
        region=[of_type(six.text_type)],
        number_type=[optional(of_type(six.text_type))],
        app_id=[optional(of_type(six.text_type))],
        subaccount=[optional(is_subaccount())])
    def create(self,
               numbers,
               carrier,
               region,
               number_type=None,
               app_id=None,
               subaccount=None):
        return self.client.request('POST', ('Number', ),
                                   to_param_dict(self.create, locals()))

    def update(self,
               number,
               app_id=None,
               subaccount=None,
               alias=None,
               verification_info=None,):
        return self.client.request('POST', ('Number', number),
                                   to_param_dict(self.update, locals()))

    def delete(self, number):
        return self.client.request('DELETE', ('Number', number))



from plivo.base import (PlivoResource, PlivoResourceInterface)


class EndUser(PlivoResource):
    _name = 'EndUser'
    _identifier_string = 'end_user'


class EndUsers(PlivoResourceInterface):
    def __init__(self, client):
        self._resource_type = EndUser
        super(EndUsers, self).__init__(client)

    def get(self, end_user_id):
        return self.client.request('GET', ('EndUser', end_user_id), response_type=EndUser)

    def list(self, limit=None, offset=None, name=None, last_name=None, end_user_type=None):
        return self.client.request('GET', ('EndUser',), dict(limit=limit, offset=offset,
                                                             name=name, last_name=last_name, end_user_type=end_user_type))

    def create(self, name=None, last_name=None, end_user_type=None):
        return self.client.request('POST', ('EndUser',),
                                   dict(name=name, last_name=last_name, end_user_type=end_user_type))

    def update(self, end_user_id=None, last_name=None, name=None, end_user_type=None):
        return self.client.request('POST', ('EndUser', end_user_id),
                                   dict(name=name, last_name=last_name, end_user_type=end_user_type))

    def delete(self, end_user_id=None):
        return self.client.request('DELETE', ('EndUser', end_user_id))


class ComplianceDocumentType(PlivoResource):
    _name = 'ComplianceDocumentType'
    _identifier_string = 'compliance_document_type'


class ComplianceDocumentTypes(PlivoResourceInterface):
    def __init__(self, client):
        self._resource_type = EndUser
        super(ComplianceDocumentTypes, self).__init__(client)

    def get(self, doc_id):
        return self.client.request('GET', ('ComplianceDocumentType', doc_id), response_type=ComplianceDocumentType)

    def list(self, limit=None, offset=None):
        return self.client.request('GET', ('ComplianceDocumentType',), dict(limit=limit, offset=offset))


class ComplianceDocument(PlivoResource):
    _name = 'ComplianceDocument'
    _identifier_string = 'compliance_document'


class ComplianceDocuments(PlivoResourceInterface):
    def __init__(self, client):
        self._resource_type = EndUser
        super(ComplianceDocuments, self).__init__(client)

    def get(self, compliance_document_id):
        return self.client.request('GET', ('ComplianceDocument', compliance_document_id),
                                   response_type=ComplianceDocument)

    def list(self, limit=None, offset=None, end_user_id=None, document_type_id=None, alias=None):
        return self.client.request('GET', ('ComplianceDocument',), dict(limit=limit, offset=offset,
                                                                        end_user_id=end_user_id,
                                                                        document_type_id=document_type_id, alias=alias))

    def create(self, end_user_id=None, document_type_id=None, alias=None, file=None, **data_fields):
        payload, files = construct_compliance_document_payload(end_user_id, document_type_id, alias, file, data_fields)
        return self.client.request('POST', ('ComplianceDocument',), payload, files=files)

    def update(self, compliance_document_id=None, end_user_id=None, document_type_id=None, alias=None, file=None,
               **data_fields):
        payload, files = construct_compliance_document_payload(end_user_id, document_type_id, alias, file, data_fields)
        return self.client.request('POST', ('ComplianceDocument', compliance_document_id), payload, files=files)

    def delete(self, compliance_document_id=None):
        return self.client.request('DELETE', ('ComplianceDocument', compliance_document_id))


class ComplianceRequirement(PlivoResource):
    _name = 'ComplianceRequirement'
    _identifier_string = 'compliance_requirement'


class ComplianceRequirements(PlivoResourceInterface):
    def __init__(self, client):
        self._resource_type = ComplianceRequirement
        super(ComplianceRequirements, self).__init__(client)

    def get(self, compliance_requirement_id):
        return self.client.request('GET', ('ComplianceRequirement', compliance_requirement_id),
                                   response_type=ComplianceRequirement)

    def list(self, country_iso2=None, number_type=None, end_user_type=None, phone_number=None):
        return self.client.request('GET', ('ComplianceRequirement',),
                                   dict(country_iso2=country_iso2, number_type=number_type,
                                        end_user_type=end_user_type, phone_number=phone_number), )


class ComplianceApplication(PlivoResource):
    _name = 'ComplianceApplication'
    _identifier_string = 'compliance_application'


class ComplianceApplications(PlivoResourceInterface):
    def __init__(self, client):
        self._resource_type = EndUser
        super(ComplianceApplications, self).__init__(client)

    def get(self, compliance_application_id):
        return self.client.request('GET', ('ComplianceApplication', compliance_application_id),
                                   response_type=ComplianceApplication)

    def list(self, status=None, end_user_type=None, number_type=None, country_iso2=None,
             alias=None, limit=None, offset=None, end_user_id=None):
        return self.client.request('GET', ('ComplianceApplication',),
                                   dict(status=status, end_user_type=end_user_type, number_type=number_type,
                                        country_iso2=country_iso2, alias=alias, limit=limit, offset=offset,
                                        end_user_id=end_user_id), )

    def create(self, compliance_requirement_id=None, end_user_id=None, document_ids=None, alias=None,
               end_user_type=None, country_iso2=None, number_type=None):
        return self.client.request('POST', ('ComplianceApplication',),
                                   dict(compliance_requirement_id=compliance_requirement_id, end_user_id=end_user_id,
                                        alias=alias, document_ids=document_ids, end_user_type=end_user_type,
                                        country_iso2=country_iso2, number_type=number_type))

    def update(self, compliance_application_id=None, document_ids=None):
        return self.client.request('POST', ('ComplianceApplication', compliance_application_id),
                                   dict(document_ids=document_ids))

    def delete(self, compliance_application_id=None):
        return self.client.request('DELETE', ('ComplianceApplication', compliance_application_id))

    def submit(self, compliance_application_id=None):
        return self.client.request('POST', ('ComplianceApplication', compliance_application_id, 'Submit'))


# Helper methods
def construct_compliance_document_payload(end_user_id, document_type_id, alias, file, data_fields):
    payload = dict(end_user_id=end_user_id, document_type_id=document_type_id, alias=alias)
    for key, value in data_fields.items():
        payload.update({key: value})
    files = {'file': ''}
    if file:
        files = {'file': open(file, 'rb')}
    return payload, files



# -*- coding: utf-8 -*-
"""
Pricing class - along with its list class
"""

from plivo.base import PlivoResource, PlivoResourceInterface
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Pricing(PlivoResource):
    _name = 'Pricing'
    _identifier_string = 'country_iso'

    def get(self):
        return self.client.pricing.get(self.id)


class Pricings(PlivoResourceInterface):
    _resource_type = Pricing

    @validate_args(country_iso=[regex(r'^[A-Z]{2}$')])
    def get(self, country_iso):
        return self.client.request(
            'GET', ('Pricing', ),
            to_param_dict(self.get, locals()),
            response_type=Pricing)



from plivo.utils.validators import *
from plivo.xml import PlivoXMLElement, map_type


class UserElement(PlivoXMLElement):
    _name = 'User'
    _nestable = []

    @property
    def send_digits(self):
        return self.__send_digits

    @send_digits.setter
    def send_digits(self, value):
        self.__send_digits = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_send_digits(self, value):
        self.send_digits = value
        return self

    @property
    def send_on_preanswer(self):
        return self.__send_on_preanswer

    @send_on_preanswer.setter
    def send_on_preanswer(self, value):
        self.__send_on_preanswer = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_send_on_preanswer(self, value):
        self.send_on_preanswer = value
        return self

    @property
    def sip_headers(self):
        return self.__sip_headers

    @sip_headers.setter
    def sip_headers(self, value):
        self.__sip_headers = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_sip_headers(self, value):
        self.sip_headers = value
        return self

    def __init__(
            self,
            content,
            send_digits=None,
            send_on_preanswer=None,
            sip_headers=None,
    ):
        super(UserElement, self).__init__()

        self.content = content
        self.send_digits = send_digits
        self.send_on_preanswer = send_on_preanswer
        self.sip_headers = sip_headers

    def to_dict(self):
        d = {
            'sendDigits': self.send_digits,
            'sendOnPreanswer': self.send_on_preanswer,
            'sipHeaders': self.sip_headers,
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



from plivo.xml import (
    PlivoXMLElement,
    map_type,
    SpeakElement,
    PlayElement,
    WaitElement
)
from plivo.utils.validators import *


class GetInputElement(PlivoXMLElement):
    _name = 'GetInput'
    _nestable = [
        'Speak',
        'Play',
        'Wait'
    ]

    @property
    def action(self):
        return self.__action

    @action.setter
    def action(self, value):
        self.__action = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_action(self, value):
        self.action = value
        return self

    @property
    def method(self):
        return self.__method

    @method.setter
    def method(self, value):
        self.__method = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_method(self, value):
        self.method = value
        return self
    
    @property
    def input_type(self):
        return self.__input_type

    @input_type.setter
    def input_type(self, value):
        self.__input_type = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_input_type(self, value):
        self.input_type = value
        return self

    @property
    def execution_timeout(self):
        return self.__execution_timeout

    @execution_timeout.setter
    def execution_timeout(self, value):
        self.__execution_timeout = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_execution_timeout(self, value):
        self.execution_timeout = value
        return self

    @property
    def digit_end_timeout(self):
        return self.__digit_end_timeout

    @digit_end_timeout.setter
    def digit_end_timeout(self, value):
        self.__digit_end_timeout = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_digit_end_timeout(self, value):
        self.digit_end_timeout = value
        return self

    @property
    def speech_end_timeout(self):
        return self.__speech_end_timeout

    @speech_end_timeout.setter
    def speech_end_timeout(self, value):
        self.__speech_end_timeout = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_speech_end_timeout(self, value):
        self.speech_end_timeout = value
        return self

    @property
    def finish_on_key(self):
        return self.__finish_on_key

    @finish_on_key.setter
    def finish_on_key(self, value):
        self.__finish_on_key = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_finish_on_key(self, value):
        self.finish_on_key = value
        return self

    @property
    def num_digits(self):
        return self.__num_digits

    @num_digits.setter
    def num_digits(self, value):
        self.__num_digits = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_num_digits(self, value):
        self.num_digits = value
        return self

    @property
    def speech_model(self):
        return self.__speech_model

    @speech_model.setter
    def speech_model(self, value):
        self.__speech_model = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_speech_model(self, value):
        self.speech_model = value
        return self

    @property
    def hints(self):
        return self.__hints

    @hints.setter
    def hints(self, value):
        self.__hints = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_hints(self, value):
        self.hints = value
        return self

    @property
    def language(self):
        return self.__language

    @language.setter
    def language(self, value):
        self.__language = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_language(self, value):
        self.language = value
        return self

    @property
    def interim_speech_results_callback(self):
        return self.__interim_speech_results_callback

    @interim_speech_results_callback.setter
    def interim_speech_results_callback(self, value):
        self.__interim_speech_results_callback = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_interim_speech_results_callback(self, value):
        self.interim_speech_results_callback = value
        return self

    @property
    def interim_speech_results_callback_method(self):
        return self.__interim_speech_results_callback_method

    @interim_speech_results_callback_method.setter
    def interim_speech_results_callback_method(self, value):
        self.__interim_speech_results_callback_method = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_interim_speech_results_callback_method(self, value):
        self.interim_speech_results_callback_method = value
        return self

    @property
    def log(self):
        return self.__log

    @log.setter
    def log(self, value):
        self.__log = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_log(self, value):
        self.log = value
        return self

    @property
    def redirect(self):
        return self.__redirect

    @redirect.setter
    def redirect(self, value):
        self.__redirect = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_redirect(self, value):
        self.redirect = value
        return self

    @property
    def profanity_filter(self):
        return self.__profanity_filter

    @profanity_filter.setter
    def profanity_filter(self, value):
        self.__profanity_filter = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_profanity_filter(self, value):
        self.profanity_filter = value
        return self

    def __init__(
            self,
            action=None,
            method=None,
            input_type=None,
            execution_timeout=None,
            digit_end_timeout=None,
            speech_end_timeout=None,
            finish_on_key=None,
            num_digits=None,
            speech_model=None,
            hints=None,
            language=None,
            interim_speech_results_callback=None,
            interim_speech_results_callback_method=None,
            log=None,
            redirect=None,
            profanity_filter=None,
    ):
        super(GetInputElement, self).__init__()

        self.action = action
        self.method = method
        self.input_type = input_type
        self.execution_timeout = execution_timeout
        self.digit_end_timeout = digit_end_timeout
        self.speech_end_timeout = speech_end_timeout
        self.finish_on_key = finish_on_key
        self.num_digits = num_digits
        self.speech_model = speech_model
        self.hints = hints
        self.language = language
        self.interim_speech_results_callback = interim_speech_results_callback
        self.interim_speech_results_callback_method = interim_speech_results_callback_method
        self.log = log
        self.redirect = redirect
        self.profanity_filter = profanity_filter

    def to_dict(self):
        d = {
            'action': self.action,
            'method': self.method,
            'inputType': self.input_type,
            'executionTimeout': self.execution_timeout,
            'digitEndTimeout': self.digit_end_timeout,
            'speechEndTimeout': self.speech_end_timeout,
            'finishOnKey': self.finish_on_key,
            'numDigits': self.num_digits,
            'speechModel': self.speech_model,
            'hints': self.hints,
            'language': self.language,
            'interimSpeechResultsCallback': self.interim_speech_results_callback,
            'interimSpeechResultsCallbackMethod': self.interim_speech_results_callback_method,
            'log': self.log,
            'redirect': self.redirect,
            'profanityFilter': self.profanity_filter
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }

    def add_speak(
            self,
            content,
            voice=None,
            language=None,
            loop=None,
    ):
        self.add(
            SpeakElement(
                content=content,
                voice=voice,
                language=language,
                loop=loop,
            ))
        return self

    def add_play(
            self,
            content,
            loop=None,
    ):
        self.add(PlayElement(
            content=content,
            loop=loop,
        ))
        return self

    def add_wait(
            self,
            length=None,
            silence=None,
            min_silence=None,
            beep=None,
    ):

        self.add(
            WaitElement(
                length=length,
                silence=silence,
                min_silence=min_silence,
                beep=beep,
            ))
        return self
